-- A3Pv2.8/proasic3v3/VectGen.vhdl
-- created dim. déc.  8 08:36:48 CET 2019 by whygee@f-cpu.org
-- version dim. déc.  8 22:01:42 CET 2019 : gets the DUT's vector size.
-- version sam. déc. 21 09:52:10 CET 2019 : depth
-- version mar. déc. 24 22:25:32 CET 2019 : GateList(g).depth_max fixed
--
-- This program extracts and analyses a netlist built with the A3P gates.
-- This netlist provides useful information about the circuit, it might find some potential errors
-- and more importantly : it will build test vectors that could be used later to verify
-- the integrity and functionality of a physical implementation.
-- It is part of the A3P library and one major reason for its evolution.
--
-- Released under the GNU AGPLv3 license

Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
use STD.textio.all;
Library SLV;
    use SLV.SLV_utils.all;
Library Histo;
    use Histo.Histograms.all;
Library work;
    use work.all;
Library proasic3;
    use proasic3.all;
    use proasic3.PA3_genlib.all;

entity VectGen is
  generic (
    VectGenWidthIn : integer := 200;
    VectGenWidthOut: integer := 100;
    main_delay : time := 1 sec;
    filename : string := ""; -- file containing the exclude vectors
    verbose : string := ""
  );
end VectGen;

architecture plip of VectGen is
  signal VectIn : std_logic_vector(VectGenWidthIn -1 downto 0) := (others=>'U');
  signal VectOut: std_logic_vector(VectGenWidthOut-1 downto 0) := (others=>'U');
  signal VectClk: std_logic := '0';

  shared variable VectInMax  : integer := VectGenWidthIn -1;
  shared variable VectOutMax : integer := VectGenWidthOut-2; -- -1 is used by VectInMax loopback

  -- depth linked list: stores variable arrays of gate ids
  type depth_ll_record;
  type DepthRecord_access is access depth_ll_record;
  type depth_ll_record is record
    list : integer_array_ptr;
    count : integer;
    NextDepthRecord : DepthRecord_access;
  end record;
  shared variable FirstDepthItem: DepthRecord_access;
  type DepthRecord_axarray is array(integer range <>) of DepthRecord_access;
  type DepthRecord_axaxarray is access DepthRecord_axarray;
  shared variable DepthList: DepthRecord_axaxarray;
  shared variable Depth, valid_Depth: integer;

  shared variable Zombie_gates: integer_array_ptr;
  shared variable Zombies_count: integer;

begin

  -- force the registration of the gates when placed at the very beginning
  rg: entity register_generics generic map(
         gate_select_number => 1, -- no listing
         bit_select_number => -1, -- no alteration
         flip_LUT => 2,           -- probe mode
         filename => filename,    -- eventual exclusions
         verbose => verbose)
  port map(clock => '0');   -- the updates are directly driven by main

  -- This code is executed after all the variables/signals are initialised
  main: process

------------------------------------------------------------------------------------------

    procedure GetVectorsSize is
    begin
      wait for main_delay;  -- let the circuit settle so we can read the output

      -- Find the size of the output vector
      FindOut: loop
        assert VectOutMax > 1
             report " Can't find the output vector size :-( "
             severity failure;
        exit when VectOut(VectOutMax) = '1';
        VectOutMax := VectOutMax - 1;
      end loop;

      -- Then find the size of the input vector
      FindIn: loop
        assert VectInMax > 1
          report " Can't find the intput vector size :-( "
          severity failure;
        VectIn(VectInMax) <= '1';
        wait for main_delay;
        exit when VectOut(VectGenWidthOut-1) = '1';
        VectInMax := VectInMax - 1;
      end loop;

      if verbose /= "" then
        report " Input vector : " & integer'image(VectInMax)
            & " bits, Output vector : " & integer'image(VectOutMax) & " bits";
      end if;

      -- avoid off-by-one
      VectInMax := VectInMax - 1;
      VectOutMax := VectOutMax - 1;

      -- allocate some memory for the descriptors
      output_records := new input_records_array(0 to VectOutMax);
       input_records := new output_record_array(0 to VectInMax);

    end GetVectorsSize;

------------------------------------------------------------------------------------------

    procedure GetNetList is
      variable i : integer;
      variable j : std_logic;

      procedure check_inputs is
      begin
        if verbose = "full" then
          report " \\\\\\\\\\\ BATCH TEST \\\\\\\\\\\ ";
        end if;

        wait for main_delay;

        -- check the outputs too !
        for i in 0 to VectOutMax loop
          j := VectOut(i);
          if (j /= 'U') and (j /= 'X') and
           output_records(i).source = integer'low then
            output_records(i) := register_source(-i, -1, j);
          end if;
        end loop;

        -- might no be so useless after all...
        probe_phase := 'X';
        VectIn(VectInMax downto 0) <= (others => probe_phase);
        wait for main_delay;

        -- erase the previous input vectors
        probe_phase := 'U';
        VectIn(VectInMax downto 0) <= (others => probe_phase);

        if verbose = "full" then
          report "////////  BATCH TEST END //////// ";
        end if;
      end check_inputs;


    begin
      probe_phase := 'X';
      VectIn(VectInMax downto 0) <= (others => probe_phase);
      wait for main_delay;

      probe_phase := 'U';
      VectIn(VectInMax downto 0) <= (others => probe_phase);

      j := '0';
      probe_numbers := (others => integer'low);

      -- scan the inputs first :
      for i in 0 to VectInMax loop
        if verbose = "full" then
          report " Toggling input #" & integer'image(i) & "   _  " & std_logic'image(j);
        end if;
        probe_numbers(j) := -i;  -- 0=>0, 1=>-1, 2=>-2...
        VectIn(i) <= j;
        if j = '-' then
          check_inputs;
          probe_numbers := (others => integer'low);
          j := '0';
        else
          j := std_logic'succ(j);
        end if;
      end loop;

      -- now scan the gates !
      for i in 1 to gate_instance_counter loop
        if verbose = "full" then
          report " Toggling output of gate #" & integer'image(i) & "   _  " & std_logic'image(j);
        end if;
        probe_numbers(j) := i;  -- i >= 1

        if j = '-' then
          check_inputs;
          probe_numbers := (others => integer'low);
          j := '0';
        else
          j := std_logic'succ(j);
        end if;
      end loop;

      if j /= '0' then
        check_inputs;
      end if;

    end GetNetList;

------------------------------------------------------------------------------------------

    procedure FixSinklist is
      variable i, j : integer;

      procedure compile_sinks(driver: inout output_record_type) is
        variable g, n, o, p, g2, p2 : integer;
      begin
        g := driver.first_sink;
        n := driver.fanout;
        o := 1;
        p := driver.first_port;
        driver.sinks_array := new input_address_array(1 to n);

        loop
          driver.sinks_array(o) := (g, p);
          if g > 0 then
            if verbose = "full" then
              report "     " & integer'image(o) & " : Gate #" & integer'image(g)
                   & "(" & integer'image(p) & ") - " & trim_name(GateList(g).gatename.all, "_wrapper(wrap_");
            end if;
            g2 := GateList(g).in_record(p).next_input;
            p2 := GateList(g).in_record(p).next_port;
          else
            if verbose = "full" then            
              report "     " & integer'image(o) & " : Output #" & integer'image(-g);
            end if;
            g2 := output_records(-g).next_input;
            p2 := output_records(-g).next_port;
          end if;
          n := n-1;
          exit when n < 1;
          o := o+1;
          g := g2;
          p := p2;
        end loop;
      end compile_sinks;

    begin
      if verbose /= "" then
        report " ************ FIXING NETLIST ************";
      end if;

      -- scan the inputs first :
      for i in 0 to VectInMax loop
        j := input_records(i).fanout;
        if j > 0 then
          if verbose = "full" then
            report "Fixing Input #" & integer'image(i) & " : fanout=" & integer'image(j);
          end if;
          compile_sinks(input_records(i));
        else
          report "Warning : Input #" & integer'image(i) & " : NO SINK / OPEN";
        end if;
      end loop;

      -- scan the gates
      for i in 1 to gate_instance_counter loop
        j := GateList(i).out_record.fanout;
        if j > 0 then
          if verbose = "full" then
            report "Fixing Gate #" & integer'image(i) & " : fanout=" & integer'image(j);
          end if;
          compile_sinks(GateList(i).out_record);
        else
          report "- Gate #" & integer'image(i) & " : NO SINK / OPEN";
	end if;
      end loop;

      if verbose /= "" then
        report "Netlist : fixed";
      end if;
    end FixSinklist;


------------------------------------------------------------------------------------------

    procedure FixDepthList is
      variable i, j, k, current_depth, registered, enlisted : integer;
      variable dst_list,    -- gate# to check during the next loop
               iterations,  -- counters (0 to 3), -1 if zombie
               zombies      -- list of gate# of the zombie gates (tmp before move to Zombie_gates)
                 : integer_array(1 to gate_instance_counter);
      variable CurrentDepthRecord, LastDepthRecord, CompileDepthRecord : DepthRecord_access;

      procedure list_sinks(driver : inout input_address_access) is
        variable g, h, l, p : integer;
      begin
	for h in driver.all'range loop
          g := driver(h).gate_number;
          if g > 0 then
            p:=driver(h).gate_port;
            if verbose = "full" then
              report "found Gate #" & integer'image(g) & "(" & integer'image(p) & ")";
            end if;

            assert GateList(g).in_record(p).depth = integer'low
               report "Depth already initialised ???" severity failure;
            GateList(g).in_record(p).depth := current_depth;

            if GateList(g).depth_min  = integer'low then
               GateList(g).depth_min := current_depth;
              -- Insert "first time ?" meme here.
            end if;

            l := iterations(g)  - 1;
                 iterations(g) := l;

            if l = 0 then
              -- all inputs have been set so we register that gate for the next turn.
              if verbose /= "" then
                report "  > registering Gate #" & integer'image(g);
              end if;
              enlisted := enlisted + 1;
              dst_list(enlisted) := g;
              -- finally set the depth so we know the gate is good.
              GateList(g).depth_max := current_depth;
              valid_Depth := valid_Depth-1;
            end if;

          else
            if verbose /= "" then
              report "found Output #" & integer'image(-g);
            end if;
            output_records(-g).depth := current_depth;
          end if;
        end loop;
      end list_sinks;

      procedure register_zombie(i : integer) is
        -- i can be a gate or an input
      begin
        if i > 0 then
          -- don't over-count a zombie
          if iterations(i) >= 0 then
             iterations(i) := -1;
            Zombies_count := Zombies_count+1;
            zombies(Zombies_count) := i;
            report "* Zombifying gate#" & integer'image(i) & " : "
                & trim_name(GateList(i).gatename.all, "_wrapper(wrap_");
          end if;
        else
          report "* Zombifying input#" & integer'image(-i);
        end if;
      end register_zombie;

    begin
      Zombies_count:=0;
      valid_Depth := gate_instance_counter;

      if verbose /= "" then
        report " ************ FIXING DEPTHLIST ************";
      end if;

      -- initialise the iteration counters (local copy from the GateList)
      for i in iterations'range loop
        iterations(i) := GateList(i).inputs;
      end loop;

      CurrentDepthRecord := new depth_ll_record;
      FirstDepthItem := CurrentDepthRecord;
      enlisted := 0;
      current_depth := 1;
      if verbose /= "" then
        report "----- Depth=1";
      end if;

      -- scan the inputs first :
      for i in 0 to VectInMax loop
        j := input_records(i).fanout;
        if j > 0 then
          if verbose = "full" then
            report " ";
            report "- Input #" & integer'image(i) & " : fanout=" & integer'image(j);
          end if;
          list_sinks(input_records(i).sinks_array);
        else
          if verbose /= "" then
            report "Warning: Input #" & integer'image(i) & " : NO SINK / OPEN";
          end if;
          register_zombie(-i);
        end if;
      end loop;

      -- the main loop:
      loop
        CurrentDepthRecord.count := enlisted;
        exit when enlisted = 0;

        assert current_depth <= gate_instance_counter
           report "something must be terribly wrong here"
           severity failure;

        -- add the gates of this depth into a linked list record
        CurrentDepthRecord.list := new integer_array(1 to enlisted); -- allocate room
        CurrentDepthRecord.list.all := dst_list(1 to enlisted);  -- then copy the list
        LastDepthRecord := CurrentDepthRecord;
        CurrentDepthRecord := new depth_ll_record;  -- allocate a new LL entry
        LastDepthRecord.NextDepthRecord := CurrentDepthRecord; -- link the new entry to the last

        registered := enlisted;
        enlisted := 0;
        current_depth := current_depth + 1;
        if verbose /= "" then
          report "----- Depth=" & integer'image(current_depth);
        end if;

        for i in 1 to registered loop
          k := LastDepthRecord.list(i);
          assert k > 0
            report "the list only contains gate numbers, so why did I find an output here ?"
            severity failure;
          j := GateList(k).out_record.fanout;
          if j > 0 then
            if verbose = "full" then
              report " ";
              report "- Gate #" & integer'image(i) & " : fanout=" & integer'image(j)
                 & """ - " & trim_name(GateList(i).gatename.all, "_wrapper(wrap_");
            end if;
            list_sinks(GateList(k).out_record.sinks_array);
          else
            if verbose = "full" then
              report " ";
              report "No output for Gate #" & integer'image(i);
            end if;
            register_zombie(k);
          end if;
        end loop;
      end loop;

      Depth := current_depth-1;

      -- sanity check and first census of the undead gates:
      for i in iterations'range loop
        j := iterations(i);
        if j > 0 then
          if verbose = "full" then
            report "- Gate #" & integer'image(i) & " : unconnected inputs or loop ? "
                    & trim_name(GateList(i).gatename.all, "_wrapper(wrap_");
          end if;
          register_zombie(i);
        end if;
      end loop;

      -- check the outputs as well :
      for i in output_records'range loop
        j := output_records(i).depth;
        assert j > 0
           report "Output#" & integer'image(i) &  " : depth is not defined !"
           severity warning;
-- Scan the height from the outputs
      end loop;

      -- save the compacted list of zombie gates:
      Zombie_gates := new integer_array(1 to Zombies_count);
      Zombie_gates.all := zombies(1 to Zombies_count);

      -- re-compile the linked depthlist into an array
      report "Depth = " & integer'image(Depth) & " logic layers"; 
      DepthList := new DepthRecord_axarray(1 to current_depth);
      CompileDepthRecord := FirstDepthItem;
      i := 1;

      loop
        enlisted := CompileDepthRecord.count;
        DepthList(i) := CompileDepthRecord;
        exit when enlisted = 0;
        i := i+1;
        CompileDepthRecord := CompileDepthRecord.NextDepthRecord;
      end loop;

      if verbose /= "" then
        report "DepthList : fixed";
      end if;
    end FixDepthList;


------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------
-- Obsolete :

    procedure DisplayNetList is
      variable i, j : integer;
      variable h : Histogram_record;

      procedure list_sinks(driver : inout input_address_access) is
        variable g, k, p : integer;
      begin
        update_histogram(driver.all'high, h);
	for k in driver.all'range loop
          g := driver(k).gate_number;
          p := driver(k).gate_port;
          if g > 0 then
            report "     " & integer'image(k) & " : Gate #" & integer'image(g)
                 & "(" & integer'image(p) & ") - " & trim_name(GateList(g).gatename.all, "_wrapper(wrap_");
          else
            report "     " & integer'image(k) & " : Output #" & integer'image(-g);
          end if;
        end loop;
      end list_sinks;

    begin
      report " ************ NETLIST ************";
      h := create_histogram(0, 100);

      -- scan the inputs first :
      for i in 0 to VectInMax loop
        j := input_records(i).fanout;
        if j > 0 then
          report " ";
          report "- Input #" & integer'image(i) & " : fanout=" & integer'image(j);
          list_sinks(input_records(i).sinks_array);
        else
          report "- Input #" & integer'image(i) & " : NO SINK / OPEN";
        end if;
      end loop;

      -- scan the gates
      for i in 1 to gate_instance_counter loop
        j := GateList(i).out_record.fanout;
        if j > 0 then
          report " ";
          report "- Gate #" & integer'image(i) & " : fanout=" & integer'image(j)
               & "  Depth min=" & integer'image(GateList(i).depth_min)
               & "  max=" & integer'image(GateList(i).depth_max)
               & "  LUT=""" & SLV_to_bin(GateList(i).LUT(0 to GateList(i).last))
               & """ - " & trim_name(GateList(i).gatename.all, "_wrapper(wrap_");
          list_sinks(GateList(i).out_record.sinks_array);
        else
          report "- Gate #" & integer'image(i) & " : NO SINK / OPEN  - "
            & trim_name(GateList(i).gatename.all, "_wrapper(wrap_");
	end if;
      end loop;

      -- scan the outputs ?

      report " ************ END OF NETLIST ************";
      report " ";
      report " Fanout:  Count:  .........|.........|.........|.........|.........|";
      display_histogram(h, 50);
      deallocate(h.bins);

    end DisplayNetList;

------------------------------------------------------------------------------------------
-- Preferred :

    procedure DisplayDepthList is
      variable d, h : Histogram_record;
      variable i, j, l, m, q : integer;
      variable n : DepthRecord_access;
      variable o : integer_array_ptr;

      procedure list_sinks(driver : inout input_address_access) is
        variable g, k, p : integer;
      begin
        update_histogram(l, d);
	update_histogram(driver.all'high, h);
        for k in driver.all'range loop
          g := driver(k).gate_number;
          p := driver(k).gate_port;
          if g > 0 then
            report "     " & integer'image(k) & " : Gate #" & integer'image(g)
                 & "(" & integer'image(p) & ") - " & trim_name(GateList(g).gatename.all, "_wrapper(wrap_");
          else
            report "     " & integer'image(k) & " : Output #" & integer'image(-g);
          end if;
        end loop;
      end list_sinks;

    begin
      report " ************ DEPTHLIST ************";

      d := create_histogram(0, Depth);
      h := create_histogram(0, 200);

      -- scan the inputs first :
      report "";
      report " Depth 0 : "& integer'image(VectInMax+1) &" inputs.";

      l := 0;
      for i in 0 to VectInMax loop
        j := input_records(i).fanout;
        if j > 0 then
          report "- Input #" & integer'image(i) & " : fanout=" & integer'image(j);
          list_sinks(input_records(i).sinks_array);
        else
          report "- Input #" & integer'image(i) & " : NO SINK / OPEN";
        end if;
      end loop;

      -- scan the depthlist:
      loop
        l := l+1;
        exit when l > Depth;

        n := DepthList(l);
        m := n.count;
        o := n.list;
        report "";
        report "Depth " & integer'image(l) & " : " & integer'image(m) & " gates.";
        for i in o'range loop
          q := o(i);
          j := GateList(q).out_record.fanout;
          if j > 0 then
            report "- Gate #" & integer'image(q) & " : fanout=" & integer'image(j)
                 & "  Depth min=" & integer'image(GateList(q).depth_min)
                 & "  max=" & integer'image(GateList(q).depth_max)
                 & "  LUT=""" & SLV_to_bin(GateList(q).LUT(0 to GateList(q).last))
                 & """ - " & trim_name(GateList(q).gatename.all, "_wrapper(wrap_");
            list_sinks(GateList(q).out_record.sinks_array);
          else
            report "- Gate #" & integer'image(q) & " : NO SINK / OPEN  - "
              & trim_name(GateList(q).gatename.all, "_wrapper(wrap_");
          end if;
        end loop;
      end loop;

      -- output latency :
      report "";
      report " Latency of the " & integer'image(output_records'length) & " outputs :";
      for i in output_records'range loop
        j := output_records(i).depth;
        if j < 0 then
          report "     Output#" & integer'image(i) & " : N/A";
        else
          report "     Output#" & integer'image(i) & " : " & integer'image(output_records(i).depth);
        end if;
      end loop;

      -- List of zombies :
      if Zombies_count > 0 then
        report "";
        report " Found " & integer'image(Zombies_count) & " zombie gates or inputs (unconnected or loops) :";
        for i in Zombie_gates'range loop
          j := Zombie_gates(i);
          if j > 0 then
            l := GateList(j).out_record.fanout;
            if l = integer'low then
              report "- Gate #" & integer'image(j) & " : fanout= N/A - "
                   & trim_name(GateList(j).gatename.all, "_wrapper(wrap_");
            else
              report "- Gate #" & integer'image(j) & " : fanout=" & integer'image(l)
                   & " - " & trim_name(GateList(j).gatename.all, "_wrapper(wrap_");
            end if;
          else
            report "- Input #" & integer'image(j) & " : NO SINK / OPEN";
          end if;
        end loop;
      end if;

      report " ";
      report " ************ END OF DEPTHLIST ************";
      report " ";
      report " Fanout:  Count:  .........|.........|.........|.........|.........|";
      display_histogram(h, 50);
      deallocate(h.bins);

      report " ";
      report "  Depth:  Gates:  .........|.........|.........|.........|.........|";
      display_histogram(d, 50);
      deallocate(d.bins);
    end DisplayDepthList;

------------------------------------------------------------------------------------------

  begin
    GetVectorsSize;
    GetNetList;
    FixSinklist;
    FixDepthList;
-- FixAllInputs;
--    DisplayNetList;
    DisplayDepthList;
    wait;
  end process;

  -- Here we "plug" the wrapper to the Vector Generator ports.
  -- It must be the last section of the code so the drivers
  -- can initialise all the data behind the scene.
  wrap: entity VG_Wrapper
    generic map(
      VectGenWidthIn  => VectGenWidthIn,
      VectGenWidthOut => VectGenWidthOut)
    port map (
      VectIn  => VectIn,
      VectOut => VectOut,
      VectClk => VectClk);

end plip;
